'use strict';

var randomId = require('alphanumeric-id');
var ndjson = require('ndjson');
var duplexer = require('duplexer3');

var _require = require('events'),
    EventEmitter = _require.EventEmitter;

var HANDSHAKE = 0; // to find the leader
var HAVE = 1; // peer has an item
var RECEIVE = 2; // leader is going to send an item
var SEND = 3; // leader is going to wait for an item

var createPeer = function createPeer(onItem) {
	var have = Object.create(null);
	var want = [];

	var add = function add(item) {
		var id = randomId(12);
		have[id] = item;
		peer.emit('_have', id);
	};

	var all = function all() {
		var all = [];
		for (var id in have) {
			all.push(have[id]);
		}return all;
	};

	var replicate = function replicate() {
		var handshakeDone = false,
		    isLeader = void 0,
		    x = void 0;

		var peerHas = [];
		var receivingId = null;

		var onPkt = function onPkt(pkt) {
			// [command, optional payload]
			if (receivingId) {
				if (!(receivingId in have)) {
					have[receivingId] = pkt;
					peer.emit('_have', receivingId);
					peer.emit('add', pkt);
				}
				receivingId = null;

				setImmediate(onItem, pkt);
				if (isLeader) setTimeout(tick);
				return;
			}

			if (!Array.isArray(pkt) || pkt.length === 0) {
				return cb(new Error('invalid packet format'));
			}
			var cmd = pkt[0];

			if (cmd === HANDSHAKE) {
				if (handshakeDone) return; // invalid command
				var peerX = pkt[1];
				if ('number' !== typeof peerX) return; // invalid data

				if (peerX === x) return sendHandshake();
				isLeader = peerX < x;
				handshakeDone = true;
			} else {
				var id = pkt[1];
				if ('string' !== typeof id) return; // invalid data

				if (cmd === HAVE) {
					if (!peerHas.includes(id)) peerHas.push(id);
					if (!(id in have) && !want.includes(id)) {
						want.push(id);
						peer.emit('_want', id);
					}
				} else if (cmd === RECEIVE && !isLeader) {
					receivingId = id;
				} else if (cmd === SEND && !isLeader) {
					outgoing.write(have[id]);
				} else return; // invalid command
			}

			if (isLeader && handshakeDone) setTimeout(tick);
		};

		var incoming = ndjson.parse(); // todo: multiplex?
		incoming.on('data', onPkt);
		var outgoing = ndjson.stringify(); // todo: multiplex?

		var sendHandshake = function sendHandshake() {
			x = Math.random();
			outgoing.write([HANDSHAKE, x]);
		};

		for (var id in have) {
			outgoing.write([HAVE, id]);
		}sendHandshake();
		peer.on('_have', function (id) {
			if (!peerHas.includes(id)) outgoing.write([HAVE, id]);
		});

		var tick = function tick() {
			var i = want.findIndex(function (id) {
				return peerHas.includes(id);
			});
			if (i >= 0) {
				receivingId = want[i];
				want.splice(i, 1);
				outgoing.write([SEND, receivingId]);
				return;
			}

			var id = Object.keys(have).find(function (id) {
				return !peerHas.includes(id);
			});
			if (id) {
				peerHas.push(id);
				outgoing.write([RECEIVE, id]);
				outgoing.write(have[id]);

				setTimeout(tick);
			}
		};

		return duplexer({ objectMode: true }, incoming, outgoing);
	};

	var peer = new EventEmitter();
	peer.add = add;
	peer.all = all;
	peer.replicate = replicate;
	return peer;
};

module.exports = createPeer;